r"""Gravitational waves (GW) benchmark.

The GW simulator computes the gravitational waves emitted by precessing quasi-circular
binary black hole (BBH) systems, and project them onto LIGO detectors (H1 and L1).
It assumes stationary Gaussian noise with respect to the detectors' noise spectral
densities, estimated from 1024 seconds of detector data prior to GW150914. The
waveforms are compressed to a reduced-order basis corresponding to the first 128
components of a singular value decomposition (SVD).

References:
    Observation of Gravitational Waves from a Binary Black Hole Merger
    (Abbott et al., 2016)
    https://arxiv.org/abs/1602.03837

    Complete parameter inference for GW150914 using deep learning
    (Green et al., 2021)
    https://arxiv.org/abs/2008.03312

Shapes:
    theta: :math:`(15,)`.
    x: :math:`(2, 256)`.
"""

import numpy as np
import os
import torch

try:
    os.environ['GWPY_RCPARAMS'] = '0'

    from gwpy.timeseries import TimeSeries
    from lal import MSUN_SI
    from lalsimulation import SimInspiralTransformPrecessingNewInitialConditions
    from pycbc.catalog import Merger
    from pycbc.detector import Detector
    from pycbc.psd import welch
    from pycbc.waveform import get_fd_waveform
except Exception as e:
    print(f"ImportWarning: {e}. 'GW' requires")
    print("  pip install gwpy pycbc")

from numpy import ndarray as Array
from torch import Tensor, BoolTensor
from tqdm import tqdm
from typing import *

from . import Simulator
from ..distributions import (
    Distribution,
    Joint,
    Uniform,
    Sort,
    Maximum,
    Minimum,
    TransformedUniform,
    CosTransform,
    SinTransform,
)
from ..utils import cache, vectorize


LABELS = [
    f'${l}$' for l in [
        'm_1', 'm_2', r'\phi_c', 't_c', 'd_L',
        'a_1', 'a_2', r'\theta_1', r'\theta_2', r'\phi_{12}', r'\phi_{JL}',
        r'\theta_{JN}', r'\psi', r'\alpha', r'\delta',
    ]
]

LOWER, UPPER = torch.tensor([
    [10., 80.],               # primary mass [solar masses]
    [10., 80.],               # secondary mass [solar masses]
    [0., 2 * np.pi],          # coalesence phase [rad]
    [-0.1, 0.1],              # coalescence time [s]
    [100., 1000.],            # luminosity distance [megaparsec]
    [0., 0.88],               # a_1 [/]
    [0., 0.88],               # a_2 [/]
    [0., np.pi],              # tilt_1 [rad]
    [0., np.pi],              # tilt_2 [rad]
    [0., 2 * np.pi],          # phi_12 [rad]
    [0., 2 * np.pi],          # phi_jl [rad]
    [0., np.pi],              # theta_jn [rad]
    [0., np.pi],              # polarization [rad]
    [0., 2 * np.pi],          # right ascension [rad]
    [-np.pi / 2, np.pi / 2],  # declination [rad]
]).t()


def build_prior(b: BoolTensor = None) -> Distribution:
    r"""Returns a prior distribution :math:`p(\theta)` for BBH systems.

    Arguments:
        b: An optional binary mask :math:`b`, with shape :math:`(D,)`.
    """

    if b is None:
        b = [True] * 15

    marginals = []

    if b[0] or b[1]:
        base = Uniform(LOWER[0], UPPER[0])

        if b[0] and b[1]:
            law = Sort(base, n=2, descending=True)
        elif b[0]:
            law = Maximum(base, n=2)
        elif b[1]:
            law = Minimum(base, n=2)

        marginals.append(law)

    for i in range(2, len(b)):
        if b[i]:
            if i in [7, 8, 11]:  # [tilt_1, tilt_2, theta_jn]
                m = TransformedUniform(CosTransform(), LOWER[i], UPPER[i])
            elif i == 14:  # declination
                m = TransformedUniform(SinTransform(), LOWER[i], UPPER[i])
            else:
                m = Uniform(LOWER[i], UPPER[i])

            marginals.append(m)

    return Joint(marginals)


class GW(Simulator):
    r"""Creates a gravitational waves (GW) simulator.

    Arguments:
        reduced_basis: Whether waveform are compressed to a reduced basis or not.
        noisy: Whether noise is added to waveforms or not.
        seed: A random number generator seed.
        kwargs: Simulator settings and constants (e.g. event, approximant, ...).
    """

    def __init__(
        self,
        reduced_basis: bool = True,
        noisy: bool = True,
        seed: int = None,
        **kwargs,
    ):
        super().__init__()

        # Constants
        default = {
            'event': 'GW150914',
            'detectors': ('H1', 'L1'),
            'approximant': 'IMRPhenomPv2',
            'duration': 4.,  # s
            'buffer': 2.,  # s
            'segment': 1024.,  # s
            'sample_rate': 1024.,  # Hz
            'f_ref': 40.,  # Hz
            'f_lower': 20.,  # Hz
        }

        self.constants = {
            k: kwargs.get(k, v)
            for k, v in default.items()
        }

        self.nsd = event_nsd(**self.constants)
        self.nsd = crop_dft(self.nsd, **self.constants)

        # Reduced SVD basis
        if reduced_basis:
            self.basis = svd_basis(**self.constants)
        else:
            self.basis = None

        # RNG
        self.noisy = noisy
        self.rng = np.random.default_rng(seed)

    def __call__(self, theta: Array) -> Array:
        x = gravitational_waveform(theta, **self.constants)
        x = self.process(x)

        if self.noisy:
            x = x + self.rng.standard_normal(x.shape)

        return x

    def process(self, x: Array) -> Array:
        r"""Processes waveforms into network-friendly inputs."""

        x = crop_dft(x, **self.constants)
        x = x / self.nsd

        if self.basis is not None:
            x = x @ self.basis

        return x.view(np.float64)


@cache
def ligo_detector(name: str):
    r"""Fetches LIGO detector."""

    return Detector(name)


@cache
def event_gps(event: str = 'GW150914') -> float:
    r"""Fetches event's GPS time."""

    return Merger(event).data['GPS']


@cache
def tukey_window(
    duration: int,  # s
    sample_rate: float,  # Hz
    roll_off: float = 0.4,  # s
) -> Array:
    r"""Returns a tukey window.

    References:
        https://en.wikipedia.org/wiki/Window_function
    """

    from scipy.signal import tukey

    length = int(duration * sample_rate)
    alpha = 2 * roll_off / duration

    return tukey(length, alpha)


@cache(persist=True)
def event_nsd(
    event: str,
    detectors: Tuple[str, ...],
    duration: float,  # s
    segment: float,  # s
    **absorb,
) -> Array:
    r"""Fetches event's noise spectral density (NSD).

    Wikipedia:
        https://en.wikipedia.org/wiki/Noise_spectral_density

    References:
        https://github.com/gwastro/pycbc/blob/master/pycbc/noise/gaussian.py#L35
    """

    time = event_gps(event) - duration

    nsds = []

    for det in detectors:
        strain = TimeSeries.fetch_open_data(det, time - segment, time, cache=True).to_pycbc(copy=False)

        win = tukey_window(duration, strain.sample_rate)
        win_factor = np.sum(win ** 2) / len(win)
        psd = welch(strain, len(win), len(win), window=win, avg_method='median') * win_factor
        nsd = 0.5 * np.sqrt(psd.data / psd.delta_f)

        nsds.append(nsd)

    return np.stack(nsds)


@cache(persist=True)
def event_dft(
    event: str,
    detectors: Tuple[str, ...],
    duration: float,  # s
    buffer: float,  # s
    **absorb,
) -> Array:
    r"""Fetches event's discrete fourier transform (DFT)."""

    time = event_gps(event) + buffer

    dfts = []

    for det in detectors:
        strain = TimeSeries.fetch_open_data(det, time - duration, time, cache=True).to_pycbc(copy=False)

        win = tukey_window(duration, strain.sample_rate)
        dft = (strain * win).to_frequencyseries().cyclic_time_shift(buffer)

        dfts.append(dft.data)

    return np.stack(dfts)


@vectorize(otypes=[float] * 7)
def lal_spins(*args) -> Tuple[float, ...]:
    r"""Converts LALInference geometric parameters to LALSimulation spins."""

    return tuple(SimInspiralTransformPrecessingNewInitialConditions(*args))


@vectorize(otypes=[Array, Array])
def plus_cross(**kwargs) -> Tuple[Array, Array]:
    r"""Simulates frequency-domain plus and cross polarizations of gravitational wave."""

    hp, hc = get_fd_waveform(**kwargs)
    return hp.numpy(), hc.numpy()


def gravitational_waveform(
    theta: Array,
    event: str,
    detectors: Tuple[str, ...],
    approximant: str,
    duration: float,  # s
    sample_rate: float,  # Hz
    f_ref: float,  # Hz
    f_lower: float,  # Hz
    **absorb,
) -> Array:
    r"""Simulates a frequency-domain gravitational wave projected onto LIGO detectors.

    References:
        http://pycbc.org/pycbc/latest/html/waveform.html
        http://pycbc.org/pycbc/latest/html/detector.html
    """

    shape = theta.shape[:-1]
    theta = theta.reshape(-1, theta.shape[-1])

    # Parameters
    m_1, m_2, phi_c, t_c, d_L, a_1, a_2, tilt_1, tilt_2, phi_12, phi_jl, theta_jn, psi, alpha, delta = [
        theta[..., i] for i in range(15)
    ]

    iota, spin1x, spin1y, spin1z, spin2x, spin2y, spin2z = lal_spins(
        theta_jn, phi_jl, tilt_1, tilt_2, phi_12, a_1, a_2,
        m_1 * MSUN_SI, m_2 * MSUN_SI, f_ref, phi_c
    )

    # Gravitational wave
    hp, hc = plus_cross(
        ## Constants
        approximant=approximant,
        delta_f=1/duration,
        f_ref=f_ref,
        f_lower=f_lower,
        f_final=sample_rate/2,
        ## Variables
        mass1=m_1,
        mass2=m_2,
        coa_phase=phi_c,
        distance=d_L,
        inclination=iota,
        spin1x=spin1x,
        spin2x=spin2x,
        spin1y=spin1y,
        spin2y=spin2y,
        spin1z=spin1z,
        spin2z=spin2z,
    )

    hp, hc = np.stack(hp), np.stack(hc)

    # Projection on detectors
    time = event_gps(event)
    length = int(duration * sample_rate / 2) + 1
    angular_speeds = -1j * 2 * np.pi * np.arange(length) / duration

    strains = []

    for name in detectors:
        det = ligo_detector(name)

        ## Noiseless strain
        fp, fc = det.antenna_pattern(alpha, delta, psi, time)
        s = fp[..., None] * hp + fc[..., None] * hc

        ## Cyclic time shift
        dt = det.time_delay_from_earth_center(alpha, delta, time) + t_c
        s = s * np.exp(dt[..., None] * angular_speeds)

        strains.append(s)

    strains = np.stack(strains, axis=-2)
    strains = strains.reshape(shape + strains.shape[1:])

    return strains


def crop_dft(
    dft: Array,
    duration: float,  # s
    sample_rate: float,  # Hz
    f_lower: float,  # Hz
    **absorb,
) -> Array:
    r"""Crops low and high frequencies of discrete fourier transform (DFT)."""

    return dft[..., int(duration * f_lower):int(duration * sample_rate / 2) + 1]


@cache(persist=True)
def svd_basis(
    n_components: int = 2**7,  # 128
    n_samples: int = 2**15,  # 32768
    batch_size: int = 2**10,  # 1024
    seed: int = 0,
    **kwargs,
) -> Array:
    r"""Builds Singular Value Decompostition (SVD) basis."""

    prior = gw_prior()
    simulator = GW(reduced_basis=False, noisy=False, **kwargs)

    print("Generating samples...")

    xs = []

    for _ in tqdm(range(n_samples // batch_size), unit='sample', unit_scale=batch_size):
        theta = prior.sample((batch_size,))
        theta[..., 4] = LOWER[4]  # fixed luminosity distance
        theta = theta.numpy().astype(np.float64)

        xs.append(simulator(theta))

    x = np.stack(xs).view(np.complex128)
    x = x.reshape(-1, x.shape[-1])

    print("Computing SVD basis...")

    try:
        from sklearn.utils.extmath import randomized_svd

        _, _, Vh = randomized_svd(
            x,
            n_components=n_components,
            n_oversamples=n_components,
            random_state=seed,
        )
    except ImportError as e:
        _, _, Vh = np.linalg.svd(x, full_matrices=False)
        Vh = Vh[:n_components]

    V = Vh.T.conj()

    print("Done!")

    return V



r"""Simulators and benchmarks."""

from abc import ABC, abstractmethod
from numpy import ndarray as Array
from torch import Tensor
from typing import *


class Simulator(ABC):
    r"""Abstract simulator class."""

    @abstractmethod
    def __call__(self, theta: Union[Array, Tensor]) -> Union[Array, Tensor]:
        pass



r"""Hodgkin-Huxley (HH) benchmark.

HH is a widespread non-linear mechanistic model of neural dynamics.

References:
    A quantitative description of membrane current and its application to conduction and excitation in nerve
    (Hodgkin et al., 1952)
    https://link.springer.com/article/10.1007/BF02459568

    Training deep neural density estimators to identify mechanistic models of neural dynamics
    (Gonçalves et al., 2020)
    https://elifesciences.org/articles/56261

Shapes:
    theta: :math:`(8,)`.
    x: :math:`(7,)`.
"""

import numpy as np
import torch

from numpy import ndarray as Array
from torch import Tensor, BoolTensor
from typing import *

from . import Simulator


LABELS = [
    f'${l}$' for l in [
        r'g_{\mathrm{Na}}', r'g_{\mathrm{K}}', r'g_{\mathrm{M}}', 'g_l',
        r'\tau_{\max}', 'V_t', r'\sigma', 'E_l',
    ]
]

LOWER, UPPER = torch.tensor([
    [0.5, 80.],     # g_Na [mS/cm^2]
    [1e-4, 15.],    # g_K [mS/cm^2]
    [1e-4, .6],     # g_M [mS/cm^2]
    [1e-4, .6],     # g_l [mS/cm^2]
    [50., 3000.],   # tau_max [ms]
    [-90., -40.],   # V_t [mV]
    [1e-4, .15],    # sigma [uA/cm^2]
    [-100., -35.],  # E_l [mV]
]).t()


class HH(Simulator):
    r"""Creates an Hodgkin-Huxley (HH) simulator.

    Arguments:
        summary: Whether voltage traces are converted to summary statistics or not.
        seed: A random number generator seed.
        kwargs: Simulator settings and constants (e.g. duration, inital voltage, ...).
    """

    def __init__(self, summary: bool = True, seed: int = None, **kwargs):
        super().__init__()

        # Constants
        default = {
            'duration': 80.,  # s
            'time_step': 0.02,  # s
            'padding': 10.,  # s
            'initial_voltage': -70., # mV
            'current': 5e-4 / (np.pi * 7e-3 ** 2),  # uA / cm^2
        }

        self.constants = {
            k: kwargs.get(k, v)
            for k, v in default.items()
        }

        # Summary statistics
        self.summary = summary

        # RNG
        self.rng = np.random.default_rng(seed)

    def __call__(self, theta: Array) -> Array:
        x = voltage_trace(theta, self.constants, self.rng)

        if self.summary:
            x = summarize(x, self.constants)

        return x


def voltage_trace(
    theta: Array,
    constants: Dict[str, float],
    rng: np.random.Generator,
) -> Array:
    r"""Simulates an Hodgkin-Huxley voltage trace.

    References:
        https://github.com/mackelab/sbi/blob/main/examples/HH_helper_functions.py
    """

    # Parameters
    T = constants['duration']
    dt = constants['time_step']
    pad = constants['padding']
    V_0 = constants['initial_voltage']
    I = constants['current']

    theta = np.expand_dims(theta, axis=0)
    g_Na, g_K, g_M, g_leak, tau_max, V_t, sigma, E_leak = [
        theta[..., i] for i in range(8)
    ]

    C = 1.  # uF/cm^2
    E_Na = 53.  # mV
    E_K = -107.  # mV

    # Kinetics
    exp = np.exp
    efun = lambda x: np.where(
        np.abs(x) < 1e-4,
        1 - x / 2,
        x / (exp(x) - 1)
    )

    alpha_n = lambda x: 0.032 * efun(-0.2 * (x - 15)) / 0.2
    beta_n = lambda x: 0.5 * exp(-(x - 10) / 40)
    tau_n = lambda x: 1 / (alpha_n(x) + beta_n(x))
    n_inf = lambda x: alpha_n(x) / (alpha_n(x) + beta_n(x))

    alpha_m = lambda x: 0.32 * efun(-0.25 * (x - 13)) / 0.25
    beta_m = lambda x: 0.28 * efun(0.2 * (x - 40)) / 0.2
    tau_m = lambda x: 1 / (alpha_m(x) + beta_m(x))
    m_inf = lambda x: alpha_m(x) / (alpha_m(x) + beta_m(x))

    alpha_h = lambda x: 0.128 * exp(-(x - 17) / 18)
    beta_h = lambda x: 4 / (1 + exp(-0.2 * (x - 40)))
    tau_h = lambda x: 1 / (alpha_h(x) + beta_h(x))
    h_inf = lambda x: alpha_h(x) / (alpha_h(x) + beta_h(x))

    tau_p = lambda x: tau_max / (3.3 * exp(0.05 * (x + 35)) + exp(-0.05 * (x + 35)))
    p_inf = lambda x: 1 / (1 + exp(-0.1 * (x + 35)))

    # Iterations
    voltages = []

    V = np.full_like(V_t, V_0)
    V_rel = V - V_t

    n = n_inf(V_rel)
    m = m_inf(V_rel)
    h = h_inf(V_rel)
    p = p_inf(V)

    for t in np.arange(0, T, dt):
        tau_V = C / (
            g_Na * m**3 * h
            + g_K * n**4
            + g_M * p
            + g_leak
        )

        V_inf = tau_V * (
            E_Na * g_Na * m**3 * h
            + E_K * g_K * n**4
            + E_K * g_M * p
            + E_leak * g_leak
            + I * (pad <= t < T - pad)
            + sigma * rng.standard_normal(V.shape) / dt**0.5
        ) / C

        V = V_inf + (V - V_inf) * exp(-dt / tau_V)
        V_rel = V - V_t

        n = n_inf(V_rel) + (n - n_inf(V_rel)) * exp(-dt / tau_n(V_rel))
        m = m_inf(V_rel) + (m - m_inf(V_rel)) * exp(-dt / tau_m(V_rel))
        h = h_inf(V_rel) + (h - h_inf(V_rel)) * exp(-dt / tau_h(V_rel))
        p = p_inf(V) + (p - p_inf(V)) * exp(-dt / tau_p(V))

        voltages.append(V)

    return np.stack(voltages, axis=-1).squeeze(axis=0)


def summarize(x: Array, constants: Dict[str, float]) -> Array:
    r"""Returns summary statistics of a voltage trace."""

    # Constants
    T = constants['duration']
    dt = constants['time_step']
    pad = constants['padding']

    t = np.arange(0, T, dt)

    # Number of spikes
    spikes = np.maximum(x, -10)
    spikes = np.diff(np.sign(np.diff(spikes)))
    spikes = np.sum(spikes < 0, axis=-1)

    # Resting moments
    rest = x[..., (pad / 2 <= t) * (t < pad)]
    rest_mean = np.mean(rest, axis=-1)
    rest_std = np.std(rest, axis=-1)

    # Moments
    x = x[..., (pad <= t) * (t < T - pad)]
    x_mean = np.mean(x, axis=-1)
    x_std = np.std(x, axis=-1)

    z = (x - x_mean[..., None]) / x_std[..., None]

    x_skew = np.mean(z**3, axis=-1)
    x_kurtosis = np.mean(z**4, axis=-1)

    return np.stack([
        spikes,
        rest_mean, rest_std,
        x_mean, x_std, x_skew, x_kurtosis,
    ], axis=-1)



r"""Datasets and data loaders."""

import h5py
import numpy as np
import random
import torch

from bisect import bisect
from numpy import ndarray as Array
from pathlib import Path
from torch import Tensor, Size
from torch.distributions import Distribution
from torch.utils.data import DataLoader, Dataset, IterableDataset
from tqdm import tqdm
from typing import *


__all__ = ['JointLoader', 'H5Dataset']


class IterableJointDataset(IterableDataset):
    r"""Creates an iterable dataset of batched pairs :math:`(\theta, x)`."""

    def __init__(
        self,
        prior: Distribution,
        simulator: Callable,
        batch_shape: Size = (),
        numpy: bool = False,
    ):
        super().__init__()

        self.prior = prior
        self.simulator = simulator
        self.batch_shape = batch_shape
        self.numpy = numpy

    def __iter__(self) -> Iterator[Tuple[Tensor, Tensor]]:
        while True:
            theta = self.prior.sample(self.batch_shape)

            if self.numpy:
                x = self.simulator(theta.detach().cpu().numpy().astype(np.float64))
                x = torch.from_numpy(x).to(theta)
            else:
                x = self.simulator(theta)

            yield theta, x


def JointLoader(
    prior: Distribution,
    simulator: Callable,
    batch_size: int = 2**10,  # 1024
    batched: bool = False,
    numpy: bool = False,
    **kwargs,
) -> DataLoader:
    r"""Creates a data loader of batched pairs :math:`(\theta, x)` generated by
    a prior distribution :math:`p(\theta)` and a simulator.

    The simlator is a stochastic function taking a set of parameters :math:`\theta`
    as input and returning an observation :math:`x` as output, which (implicitely)
    defines a likelihood function :math:`p(x | \theta)`. Together with the prior,
    they form a joint distribution :math:`p(\theta, x) = p(\theta) p(x | \theta)`
    from which the pairs :math:`(\theta, x)` are independently drawn.

    Arguments:
        prior: A prior distribution :math:`p(\theta)`.
        simulator: A callable simulator.
        batch_size: The batch size of the generated pairs.
        batched: Whether the simulator accepts batched inputs or not.
        numpy: Whether the simulator requires NumPy or PyTorch inputs.
        kwargs: Keyword arguments passed to :class:`torch.utils.data.DataLoader`.

    Returns:
        An infinite data loader of batched pairs :math:`(\theta, x)`.

    Example:
        >>> loader = joint_loader(prior, simulator, numpy=True, num_workers=4)
        >>> for theta, x in loader:
        ...     theta, x = theta.cuda(), x.cuda()
        ...     something(theta, x)
    """

    return DataLoader(
        IterableJointDataset(
            prior,
            simulator,
            batch_shape=(batch_size,) if batched else (),
            numpy=numpy,
        ),
        batch_size=None if batched else batch_size,
        **kwargs,
    )


class H5Dataset(object):
    r"""Creates a dataset of pairs :math:`(\theta, x)` from HDF5 files.

    As a :class:`torch.utils.data.Dataset`, :class:`H5Dataset` implements the methods
    :meth:`__len__` and :meth:`__getitem__`. However, as it can be slow to load pairs
    from disk one by one when iterating over the dataset, it also implements a custom
    :meth:`__iter__` method. This method loads several contiguous chunks of pairs at
    once, concatenates them, shuffles the result and, finally, splits it into batches.
    This "weak shuffling" procedure greatly improves loading performances, but the
    resulting batch elements are not perfectly independent from each others.

    Important:
        To take advantage of the custom :meth:`__iter__` method, :class:`H5Dataset`
        instances should not be wrapped inside a :class:`torch.utils.data.DataLoader`
        when iterating over the dataset.

    Arguments:
        files: HDF5 files containing pairs :math:`(\theta, x)`.
        batch_size: The size of the batches.
        chunk_size: The size of the contiguous chunks.
        group_size: The number of chunks loaded at once.
        pin_memory: Whether the batches reside in CUDA pinned memory or not.
        shuffle: Whether the pairs are shuffled when iterating.
        seed: A seed to initialize the internal RNG used for shuffling.

    Example:
        >>> dataset = H5Dataset('data.h5', batch_size=256, shuffle=True)
        >>> theta, x = dataset[0]
        >>> theta
        tensor([-0.1215, -1.3641,  0.7233, -1.2150, -1.9263])
        >>> for theta, x in dataset:
        ...     theta, x = theta.cuda(), x.cuda()
        ...     something(theta, x)
    """

    def __init__(
        self,
        *files: Union[str, Path],
        batch_size: int = 2**10,  # 1024
        chunk_size: int = 2**12,  # 4096
        group_size: str = 2**4,  # 16
        pin_memory: bool = False,
        shuffle: bool = True,
        seed: int = None,
    ):
        super().__init__()

        self.files = list(map(h5py.File, files))
        self.stops = np.cumsum([len(f['x']) for f in self.files])

        self.batch_size = batch_size
        self.chunk_size = chunk_size
        self.group_size = group_size

        self.pin_memory = pin_memory

        self.shuffle = shuffle
        self.rng = np.random.default_rng(seed)

    def __del__(self) -> None:
        for f in self.files:
            f.close()

    def __len__(self) -> int:
        return self.stops[-1]

    def __getitem__(self, i: int) -> Tuple[Tensor, Tensor]:
        i = i % len(self)
        j = bisect(self.stops, i)
        if j > 0:
            i = i - self.stops[j - 1]

        f = self.files[j]

        if 'theta' in f:
            theta = torch.from_numpy(f['theta'][i])
        else:
            theta = None

        x = torch.from_numpy(f['x'][i])

        return theta, x

    def __iter__(self) -> Iterator[Tuple[Tensor, Tensor]]:
        chunks = [
            (i, j, j + self.chunk_size)
            for i, f in enumerate(self.files)
            for j in range(0, len(f['x']), self.chunk_size)
        ]

        if self.shuffle:
            self.rng.shuffle(chunks)

        for l in range(0, len(chunks), self.group_size):
            slices = sorted(chunks[l:l+self.group_size])

            # Load
            theta = np.concatenate([self.files[i]['theta'][j:k] for i, j, k in slices])
            x = np.concatenate([self.files[i]['x'][j:k] for i, j, k in slices])

            # Shuffle
            if self.shuffle:
                order = self.rng.permutation(len(theta))
                theta, x = theta[order], x[order]

            # Tensor
            theta, x = torch.from_numpy(theta), torch.from_numpy(x)

            if self.pin_memory:
                theta, x = theta.pin_memory(), x.pin_memory()

            # Batches
            yield from zip(
                theta.split(self.batch_size),
                x.split(self.batch_size),
            )

    @staticmethod
    def store(
        pairs: Iterable[Tuple[Array, Array]],
        file: Union[str, Path],
        size: int,
        dtype: np.dtype = np.float32,
        **meta,
    ) -> None:
        r"""Creates an HDF5 file containing pairs :math:`(\theta, x)`.

        The sets of parameters :math:`\theta` are stored in a collection named
        :py:`'theta'` and the observations in a collection named :py:`'x'`.

        Arguments:
            pairs: An iterable over batched pairs :math:`(\theta, x)`.
            file: An HDF5 filename to store pairs in.
            size: The number of pairs to store.
            dtype: The data type to store pairs in.
            meta: Metadata to store in the file.

        Example:
            >>> loader = JointLoader(prior, simulator, batch_size=16)
            >>> H5Dataset.store(loader, 'sim.h5', 4096)
            100%|██████████| 4096/4096 [01:35<00:00, 42.69sample/s]
        """

        # File
        file = Path(file)
        file.parent.mkdir(parents=True, exist_ok=True)

        with h5py.File(file, 'w-') as f:
            ## Attributes
            f.attrs.update(meta)

            ## Datasets
            theta, x = map(np.asarray, next(iter(pairs)))
            theta, x = theta[0], x[0]

            f.create_dataset('theta', (size,) + theta.shape, dtype=dtype)
            f.create_dataset('x', (size,) + x.shape, dtype=dtype)

            ## Store
            with tqdm(total=size, unit='sample') as tq:
                i = 0

                for theta, x in pairs:
                    j = min(i + theta.shape[0], size)

                    f['theta'][i:j] = np.asarray(theta)[:j-i]
                    f['x'][i:j] = np.asarray(x)[:j-i]

                    tq.update(j - i)

                    if j < size:
                        i = j
                    else:
                        break



r"""Simple likelihood complex posterior (SLCP) benchmark.

SLCP is a toy simulator where :math:`\theta` parametrizes a 2-d multivariate Gaussian
from which 4 points are independently drawn and stacked as a single observation :math:`x`.
It is a non-trivial parameter inference benchmark that allows to retrieve the
ground-truth posterior through MCMC sampling of its tractable likelihood.

References:
    Sequential Neural Likelihood: Fast Likelihood-free Inference with Autoregressive Flows
    (Papamakarios et al., 2019)
    https://arxiv.org/abs/1805.07226

Shapes:
    theta: :math:`(5,)`.
    x: :math:`(8,)`.
"""

import torch

from torch import Tensor, BoolTensor
from typing import *

from . import Simulator
from ..distributions import (
    Distribution,
    Independent,
    MultivariateNormal,
    ReshapeTransform,
    TransformedDistribution,
)
from ..utils import broadcast


LABELS = [f'$\\theta_{{{i + 1}}}$' for i in range(5)]

LOWER, UPPER = torch.tensor([
    [-3., 3.],  # theta_1
    [-3., 3.],  # theta_2
    [-3., 3.],  # theta_3
    [-3., 3.],  # theta_4
    [-3., 3.],  # theta_5
]).t()


class SLCP(Simulator):
    r"""Creates a simple likelihood complex posterior (SLCP) simulator."""

    def likelihood(self, theta: Tensor, eps: float = 1e-8) -> Distribution:
        r"""Returns the likelihood distribution :math:`p(x | \theta)`."""

        # Mean
        mu = theta[..., :2]

        # Covariance
        s1 = theta[..., 2] ** 2 + eps
        s2 = theta[..., 3] ** 2 + eps
        rho = theta[..., 4].tanh()

        cov = torch.stack([
            s1 ** 2, rho * s1 * s2,
            rho * s1 * s2, s2 ** 2,
        ], dim=-1).reshape(theta.shape[:-1] + (2, 2))

        # Repeat 4 times
        mu = mu.unsqueeze(-2).repeat_interleave(4, -2)
        cov = cov.unsqueeze(-3).repeat_interleave(4, -3)

        # Normal distribution
        normal = MultivariateNormal(mu, cov)

        return TransformedDistribution(
            Independent(normal, 1),
            ReshapeTransform((4, 2), (8,)),
        )

    def __call__(self, theta: Tensor) -> Tensor:
        return self.likelihood(theta).sample()

    def log_prob(self, theta: Tensor, x: Tensor) -> Tensor:
        theta, x = broadcast(theta, x, ignore=1)
        return self.likelihood(theta).log_prob(x)



r"""General purpose helpers."""

import numpy as np
import os
import torch
import torch.nn as nn

from functools import lru_cache, partial
from itertools import islice, starmap
from torch import Tensor
from torch.optim import Optimizer
from typing import *


def cache(f: Callable = None, /, persist: bool = False) -> Callable:
    r"""Unbounded function cache decorator.

    Wraps a function with a memoizing callable that saves call results, which
    can save time when an expensive function is called several times with the
    same arguments.

    The positional and keyword arguments of :py:`f` must be hashable.

    Arguments:
        f: The function to decorate.
        persist: Whether the cached values persist to disk or not.

    Example:
        >>> @cache
        ... def fib(n):
        ...     return n if n < 2 else fib(n-2) + fib(n-1)
        ...
        >>> fib(42)
        267914296
    """

    if persist:
        try:
            from joblib import Memory
        except ImportError as e:
            print(f"ImportWarning: {e}. Fallback to regular cache.")
        else:
            memory = Memory(os.path.expanduser('~/.cache'), mmap_mode='c', verbose=0)

            if f is None:
                return memory.cache
            else:
                return memory.cache(f)

    d = lru_cache(maxsize=None)

    return d if f is None else d(f)


def vectorize(f: Callable = None, /, **kwargs):
    r"""Convenience vectorization decorator.

    Defines a vectorized function which takes a sequence of objects or NumPy arrays
    as inputs and returns a tuple of NumPy arrays. The vectorized function evaluates
    :py:`f` over successive tuples of the input arrays like the :func:`map` function,
    except it uses the broadcasting rules of NumPy.

    Arguments:
        f: The function to decorate.
        kwargs: Keyword arguments passed to :class:`numpy.vectorize`.

    Example:
        >>> @vectorize(otypes=[float])
        ... def abs(x):
        ...     return x if x > 0 else -x
        ...
        >>> abs(range(-3, 4))
        array([3., 2., 1., 0., 1., 2., 3.])
    """

    if f is None:
        return partial(vectorize, **kwargs)
    else:
        class Vectorize(np.vectorize):
            def _vectorize_call(self, func: Callable, args: List) -> Any:
                if self.signature is not None:
                    return self._vectorize_call_with_signature(func, args)
                elif not args:
                    return func()
                else:
                    ufunc, otypes = self._get_ufunc_and_otypes(func=func, args=args)

                    outputs = ufunc(*args)

                    if ufunc.nout == 1:
                        if otypes[0] == 'O':
                            return outputs
                        else:
                            return np.asanyarray(outputs, dtype=otypes[0])
                    else:
                        return tuple(
                            x if t == 'O' else np.asanyarray(x, dtype=t)
                            for x, t in zip(outputs, otypes)
                        )

        return Vectorize(f, **kwargs)


def broadcast(*tensors: Tensor, ignore: Union[int, List[int]] = None) -> List[Tensor]:
    r"""Broadcasts tensors together.

    The term broadcasting describes how PyTorch treats tensors with different shapes
    during arithmetic operations. In short, if possible, dimensions that have
    different sizes are expanded (without making copies) to be compatible.

    Arguments:
        ignore: The number(s) of dimensions not to broadcast.

    Example:
        >>> x = torch.rand(3, 1, 2)
        >>> y = torch.rand(4, 5)
        >>> x, y = broadcast(x, y, ignore=1)
        >>> x.shape
        torch.Size([3, 4, 2])
        >>> y.shape
        torch.Size([3, 4, 5])
    """

    if type(ignore) is int:
        ignore = [ignore] * len(tensors)
    dims = [t.dim() - i for t, i in zip(tensors, ignore)]

    common = torch.broadcast_shapes(*(
        t.shape[:i]
        for t, i in zip(tensors, dims)
    ))

    return [
        torch.broadcast_to(t, common + t.shape[i:])
        for t, i in zip(tensors, dims)
    ]


def deepto(obj: Any, *args, **kwargs) -> Any:
    r"""Moves and/or casts all tensors referenced in an object, recursively.

    .. warning::
        Unless a tensor already has the correct type and device, the referenced
        tensors are replaced by a copy with the desired properties, which will
        break cross-references. Proceed with caution.

    Arguments:
        obj: An object.
        args: Positional arguments passed to :func:`torch.Tensor.to`.
        kwargs: Keyword arguments passed to :func:`torch.Tensor.to`.

    Example:
        >>> tensors = [torch.arange(i) for i in range(1, 4)]
        >>> deepto(tensors, dtype=torch.float)
        [tensor([0.]), tensor([0., 1.]), tensor([0., 1., 2.])]
    """

    if torch.is_tensor(obj):
        obj = obj.to(*args, **kwargs)
    elif isinstance(obj, dict):
        for key, value in obj.items():
            obj[key] = deepto(value, *args, **kwargs)
    elif isinstance(obj, list):
        for i, value in enumerate(obj):
            obj[i] = deepto(value, *args, **kwargs)
    elif isinstance(obj, tuple):
        obj = tuple(deepto(value, *args, **kwargs) for value in obj)
    elif hasattr(obj, '__dict__'):
        deepto(obj.__dict__, *args, **kwargs)

    return obj


def starcompose(*fs: Callable) -> Callable:
    r"""Returns the composition :math:`g` of a sequence of functions
    :math:`(f_1, f_2, \dots, f_n)`.

    .. math:: g = f_n \circ \dots \circ f_2 \circ f_1

    If the output :math:`x_i` of the intermediate function :math:`f_i` is a tuple,
    its elements are used as separate arguments for the next function :math:`f_{i+1}`.

    Arguments:
        fs: A sequence of functions :math:`(f_1, f_2, \dots, f_n)`.

    Returns:
        The composition :math:`g`.

    Example:
        >>> g = starcompose(lambda x: x**2, lambda x: x/2)
        >>> g(5)
        12.5
    """

    def g(*x: Any) -> Any:
        for f in fs:
            x = f(*x) if isinstance(x, tuple) else f(x)

        return x

    return g


class GDStep(object):
    r"""Creates a callable that performs gradient descent (GD) optimization steps
    for parameters :math:`\phi` with respect to differentiable loss values.

    The callable takes a scalar loss :math:`l` as input, performs a step

    .. math:: \phi \gets \text{GD}(\phi, \nabla_{\!\phi} \, l)

    and returns the loss, detached from the computational graph. To prevent invalid
    parameters, steps are skipped if not-a-number (NaN) or infinite values are found
    in the gradient. This feature requires CPU-GPU synchronization, which could be a
    bottleneck for some applications.

    Arguments:
        optimizer: An optimizer instance (e.g. :class:`torch.optim.SGD`).
        clip: The norm at which the gradients are clipped. If :py:`None`,
            gradients are not clipped.
    """

    def __init__(self, optimizer: Optimizer, clip: float = None):

        self.optimizer = optimizer
        self.parameters = [
            p
            for group in optimizer.param_groups
            for p in group['params']
        ]
        self.clip = clip

    def __call__(self, loss: Tensor) -> Tensor:
        if loss.isfinite():
            self.optimizer.zero_grad()
            loss.backward()

            if self.clip is None:
                self.optimizer.step()
            else:
                norm = nn.utils.clip_grad_norm_(self.parameters, self.clip)
                if norm.isfinite():
                    self.optimizer.step()

        return loss.detach()


class PlateauDetector(object):
    r"""Creates a plateau detector for online sequences.

    Each time a new value :math:`x_t` is provided, it is compared with the current
    best value :math:`x_b` to determine whether :math:`t` is the new best time step.
    In the minimization mode, :math:`x_t` is considered better if it satisfies

    .. math:: x_t < x_b \, (1 - \tau) ,

    where :math:`\tau \in [0, 1]` is a significance threshold. If it is the case,
    :math:`b` becomes :math:`t`. The sequence is currently at a plateau if :math:`b`
    has not changed for more than :math:`\lambda` patience steps, i.e. if

    .. math:: t - b > \lambda .

    Arguments:
        threshold: The significance threshold :math:`\tau`.
        patience: The patience :math:`\lambda`.
        mode: The improvement mode, either :py:`'min'` or :py:`'max'`.
    """

    def __init__(
        self,
        threshold: float = 1e-2,
        patience: int = 8,
        mode: str = 'min',  # 'max'
    ):
        self.threshold = threshold
        self.patience = patience
        self.mode = mode

        self.sequence = [float('+inf' if mode == 'min' else '-inf')]
        self.best_time = 0

    @property
    def time(self) -> int:
        return len(self.sequence) - 1

    @property
    def best(self) -> float:
        return self.sequence[self.best_time]

    def step(self, value: float) -> None:
        self.sequence.append(value)

        if self.mode == 'min':
            better = value < self.best * (1 - self.threshold)
        else:
            better = value > self.best * (1 + self.threshold)

        if better:
            self.best_time = self.time

    @property
    def plateau(self) -> bool:
        return self.time - self.best_time > self.patience


def gridapply(
    self,
    f: Callable[[Tensor], Tensor],
    bins: Union[int, List[int]],
    bounds: Tuple[Tensor, Tensor],
    batch_size: int = 2**12,  # 4096
) -> Tensor:
    r"""Evaluates a function :math:`f(x)` over a multi-dimensional domain split
    into grid cells. Instead of evaluating the function cell by cell, batches are
    given to the function.

    Arguments:
        f: A function :math:`f(x)`.
        bins: The number(s) of bins per dimension.
        bounds: A tuple of lower and upper domain bounds.
        batch_size: The size of the batches given to the function.

    Example:
        >>> f = lambda x: -(x**2).sum(dim=-1) / 2
        >>> lower, upper = torch.zeros(3), torch.ones(3)
        >>> y = gridapply(f, 50, bounds=(lower, upper))
        >>> y.shape
        torch.Size([50, 50, 50])
    """

    lower, upper = bounds

    # Shape
    dims = len(lower)

    if type(bins) is int:
        bins = [bins] * dims

    # Create grid
    domains = []

    for l, u, b in zip(lower, upper, bins):
        step = (u - l) / b
        dom = torch.linspace(l, u - step, b).to(step) + step / 2.
        domains.append(dom)

    grid = torch.stack(torch.meshgrid(*domains, indexing='ij'), dim=-1)
    grid = grid.reshape(-1, dims)

    # Evaluate f(x) on grid
    y = [f(x) for x in grid.split(batch_size)]

    return torch.cat(y).reshape(*bins, *y.shape[1:])
